#!/usr/local/bin/perl
############################################################################
#  Вычисляем время, следующее за указанным, согласно шаблону в стиле CRON  #
#                          anselm.ru [2012-08-02]                          #
############################################################################

# Example: cron_next( "1-2 * * L * * L", "2012-04-04 12:32:33")
# Where: template like "min hour day month year week"

use strict;
use warnings;
use Time::Local;

print cron_next("1-2 * * L * * L", "2012-04-04 12:32:33");
#print cron_next("0 0 20 14/4 5 2012 L", "2012-04-04 12:32:33");


#print cron_next($ARGV[0], $ARGV[1]);

sub cron_next {
  my $s = shift;
  my $T = shift;
  $T = $T ? to_time($T) : time;  # опорная дата
  
  # Избавляемся от лишних пробелов
  $s =~ s/^\s+//;
  $s =~ s/\s+$//;
  $s =~ s/\s+/ /g;

  my @f = split(/ /,$s);
  $f[6] = $f[6] || '*'; # отсутствующую неделю заменяем звездой
  $f[5] = $f[5] || '*'; # отсутствующий год заменяем звездой
  
  my $t;
  if( "$f[0]$f[1]$f[2]$f[3]$f[4]$f[5]" !~ /\*/ ) { # нестандартный случай - нет звёзд
  
    $f[6] =~ /#(.+)/;
    my $week_of_month = $1 || '*'; 
    my @D = ();
    $D[7] = parse1( $week_of_month, 1,5  ); # неделя месяца
    $D[6] = parse1( $f[6], 1,7  );          # день недели
    $D[6][0] = 7 if( $D[6][0] eq 'L' );
    
    $t = find2($T, \@f);
    while($t-$T<100*366*24*3600 &&        # ищем в пределах ста лет
      (!in_arr(day_of_week($t), $D[6]) || # если день недели не соответствует - ищем дальше
        ( 
        ($D[7][0] eq 'L' && !is_last_week($t)) ^ # если неделя месяца не последняя, либо...
        ($D[7][0] ne 'L' && !in_arr(week_of_month($t), $D[7]) ) # если неделя месяца не соответствует
        )
      )
    )
    {
      $t = find2($t, \@f);
    }
    die "Error: the desired date is beyond the 100 years" if($t-$T>100*366*24*3600);
    
  }
  else {  # стандартный случай
    $t = find1($T, \@f); # найдём ближайший к точке $T узел сетки @D
  }
  
  to_str($t);
}
#####################################################################
#####################################################################
#####################################################################
# Стандартный случай
sub find1 {
  my $T = $_[0];
  my @f = @{ $_[1] };
  
  #print join(' ', @f);  exit;
  
  $f[6] =~ /#(.+)/;
  my $week_of_month = $1 || '*'; 
  
  # Разбираем шаблон
  my @t = localtime($T);
  
  my @D = ();
  $D[7] = parse1( $week_of_month, 1,5  ); # неделя месяца
  $D[6] = parse1( $f[6], 1,7  ); # день недели
  $D[6][0] = 7 if( $D[6][0] eq 'L' );
  $D[5] = parse1( $f[5], 1900+$t[5],2000+$t[5] ); # год (искать в пределах 100 лет)
  $D[4] = parse1( $f[4], 1,12 ); # месяц
  $D[3] = parse1( $f[3], 1,31 ); # день
  $D[2] = parse1( $f[2], 0,23 ); # час
  $D[1] = parse1( $f[1], 0,59 ); # минута
  $D[0] = parse1( $f[0], 0,59 ); # секунда
  
  
  # Определим начальную дату
  # Наложим граничные и начальные условия на искомую дату
  $t[5] = $D[5][0]-1900; # год
  $t[4] = $D[4][0]-1;    # месяц
  $t[3] = $D[3][0] eq 'L' ? days_of_month($t[5], $t[4]) : $D[3][0];  # день
  $t[2] = $D[2][0];  # час
  $t[1] = $D[1][0];  # минута
  $t[0] = $D[0][0];  # секунда
  #print join(' ', @{ $D[5] });  exit;
  
  
  foreach my $y (@{ $D[5] }) {
    foreach my $m (@{ $D[4] }) {
      foreach my $d (@{ $D[3] }) {
        foreach my $H (@{ $D[2] }) {
          foreach my $M (@{ $D[1] }) {
            foreach my $S (@{ $D[0] }) {
              my $dd = $d eq 'L' ? days_of_month($y-1900, $m-1) : $d;
              #print "$y\n";
              my $t; eval{ $t = timelocal($S, $M, $H, $dd, $m-1, $y-1900); };

              if( $@ =~ /^Day/ ) { # берём дней в месяце по максимуму - 31, поэтому могут быть ошибки 
                #print "NEXT_DAY ". to_str($t) ."\n" if(!$@);
                goto NEXT_DAY;
              }
              elsif( $@ )  {
                die "Error: unexpected, $@";
              }
              
              my $w = $T-$t;
              if($w>0 && int($w/3600/24/365)>0 ) {
                #print "NEXT_YEAR ". to_str($t) ."\n";
                goto NEXT_YEAR;
              }
              if($w>0 && int($w/3600/24/31)>0 ) {
                #print "NEXT_MON ". to_str($t) ."\n";
                goto NEXT_MON;
              }
              if($w>0 && int($w/3600/24)>0 ) {
                #print "NEXT_DAY ". to_str($t) ."\n";
                goto NEXT_DAY;
              }
              if(!in_arr(day_of_week($t), $D[6]) # если день недели не соответствует - пропускаем
                || ( 
                      ($D[7][0] eq 'L' && !is_last_week($t)) ^ # если неделя месяца не последняя, либо...
                      ($D[7][0] ne 'L' && !in_arr(week_of_month($t), $D[7]) ) # если неделя месяца не соответствует
                   )
              ) {
                #print "NEXT_DAY ". to_str($t) ."\n" if(!$@);
                goto NEXT_DAY;
              }
              if($w>0 && int($w/3600)>0 ) {
                #print "NEXT_HOUR ". to_str($t) ."\n";
                goto NEXT_HOUR;
              }
              if($w>0 && int($w/60)>0 ) {
                #print "NEXT_MIN ". to_str($t) ."\n";
                goto NEXT_MIN;
              }
              if($t>$T) {
                #print "FINISH   ". to_str($t) ."\n";
                return $t;
              }
            }
            NEXT_MIN:
          }
          NEXT_HOUR:
        }
        NEXT_DAY:
      }
      NEXT_MON:
    }
    NEXT_YEAR:
  }
}

#####################################################################
#####################################################################

sub parse1 {
  $_[0] =~ /^([^\/#]+)(\/|#|)(.*)$/;
  
  my $b = $3 || 1;        # период, делитель
  $b = 1 if($2 && $2 eq '#'); # если разделитель(терминатор) "#", то делитель может быть только "1"
  
  my @X = split(/,/, $1); # разбиение на числа и интервалы
  my @a = ();             # искомый массив
  foreach my $x (@X) {
    if($x =~ /^\d+$/) {
      push(@a, $x);
    }
    elsif($x =~ /^(\d+)-(\d+)$/) {
      for(my $y=$1; $y<=$2; $y++) {
        push(@a, $y);
      }
    }
  }
  
  if($X[0] eq 'L') {
    push(@a, $X[0]);
    return \@a; # массив только из одного символа
  }
  elsif($X[0] eq '*') {
    for(my $i=$_[1]; $i<=$_[2]; $i++) {
      if($i % $b == 0) {
        push(@a, $i);
      }
    }
  }

  # удалим повторяющиеся элементы массива, а так же поругаемся если пользователь задал значения вне диапазона
  my $err_msg = "Error: value out of range";
  @a = sort numerically @a;
  my @b = ();
  my $a0 = $a[0];
  die "$err_msg ($a0<$_[1] || $a0>$_[2])" if($a0<$_[1] || $a0>$_[2]);
  push(@b, $a0) ;
  for(my $i=1; $i<scalar @a; $i++) {
    die "$err_msg ($a[$i]<$_[1] || $a[$i]>$_[2])" if($a[$i]<$_[1] || $a[$i]>$_[2]);
    if($a[$i] ne $a0) {
      push(@b, $a[$i]);
    }
    $a0 = $a[$i];
  }

  #print "$_[1], $_[2]: ". join(' ', @b) ."\n";
    
  \@b;
}

#####################################################################
#####################################################################
# Поиск в случае отсутствия звёзд (недели игнорируются)
# Означает фиксированную дату плюс некий период
sub find2 {
  my $T = $_[0];
  my @f = @{ $_[1] };
  
  my @S = parse2($f[0]);
  my @M = parse2($f[1]);
  my @H = parse2($f[2]);
  my @d = parse2($f[3]);
  my @m = parse2($f[4]);
  my @y = parse2($f[5]);
  my $t; eval{ $t=timelocal($S[0],$M[0],$H[0],$d[0],$m[0]-1,$y[0]-1900); };
  if( $@ ) {
    die "Error: incorrect date ($@)";
  }
  my $b = $S[1] || $M[1]*60 || $H[1]*3600 || $d[1]*3600*24;
  #print "START    ". to_str($t) ."\n";
  
  my $a = $T-$t;
  if($a>0 && $b>0) {
    $t = $t+int($a/$b)*$b+$b;
  }
  else {
    die "Error: wrong fixed date";
  }
}

#####################################################################
#####################################################################
# Разбор №2
sub parse2 {
  $_[0] =~ /^([^\/]+)[\/]?(.*)$/;
  ($1, $2||0);
}
#####################################################################
#####################################################################

sub days_of_month { # дней в месяце
  my ($year1, $mon1, $year2, $mon2);
  if(scalar @_ == 1) {
    ($year1, $mon1) = (localtime($_[0]))[5,4];
  }
  elsif(scalar @_ == 2) {
    ($year1, $mon1) = @_;
  }
  else{
    die "Error at days_of_month - not enough arguments";
  }
  
  if($mon1<11) {
    $mon2 = $mon1+1;
    $year2 = $year1;
  }
  else {
    $mon2 = 0;
    $year2 = $year1+1;
  }
  my $t1 = timelocal(0,0,0,1,$mon1,$year1);
  my $t2 = timelocal(0,0,0,1,$mon2,$year2);
  ($t2-$t1)/(24*60*60);
}

#####################################################################
#####################################################################

sub week_of_month { # которая неделя по счёту
  my $time = shift;
  my $mon0 = (localtime($time))[4];
  my $r = 0; # искомый результат
  for($r=1; $r<5; $r++) {  # за пять циклов(недель) обязательно найдём
    $time -= 7*24*60*60;
    my $mon = (localtime($time))[4];
    last if($mon0!=$mon)
  } 
  $r;
}
#####################################################################
#####################################################################
sub is_last_week { # последний понедельник месяца, последний вторник, и т.д.
  my $time = shift;
  my $mon0 = (localtime($time))[4];
  $time += 7*24*60*60;
  my $mon = (localtime($time))[4];
  #print "$mon0==$mon\n";
  $mon0!=$mon;
}
#####################################################################
#####################################################################
sub day_of_week { # день недели
  my $i = (localtime(shift))[6];
  $i==0 ? 7 : $i;
}

#####################################################################
#####################################################################

sub is_num {
  $_[0] =~ /^\d+$/;
}
#####################################################################
#####################################################################

sub in_arr {
  my $a = $_[0];
  my @A = @{ $_[1] };
  foreach my $x (@A) {
    if($x eq $a) { # будем сравнивать не только числа, но и символы
      return 1;
    }
  }
  0;
}
#####################################################################
#####################################################################

sub numerically { # процедура для численной сортировки
  #$a*1.0 <=> $b*1.0 or $a cmp $b;
  $a <=> $b;
}
#####################################################################
#####################################################################

sub to_time { # преобразуем строку в дату
  my @t = split(/[\D]/, shift);
  timelocal($t[5]||0, $t[4]||0, $t[3]||0, $t[2], $t[1]-1, $t[0]-1900);
}

sub to_str { # преобразуем дату в строку
  my @t = (localtime(shift))[5,4,3,2,1,0];
  $t[0]+=1900;
  $t[1]++;
  sprintf("%04d-%02d-%02d %02d:%02d:%02d", @t);
}

